#include "config.h"

#include "gpgpu_frame.h"
#ifdef HAVE_CUDA
	#include "gpgpu_cuda.h"
#endif
#include "gpgpu_opencl.h"
#include <GL/glut.h>

void initGL(int, char**, const char*);

gpgpuController *gpgpuCtrl;

GLuint texHandle_gl;
int bufHandle_gpgpu;
int texHandle_gpgpu;
int kernelHandle_gpgpu;

int main(int argc, char **argv) {

	// You have to explicitly say which GPGPU backend to use
	int gpgpuMode = 0;
	for (int i = 0; i < argc; ++i) {
		if (std::string(argv[i]) == "-CUDA")
			gpgpuMode = 1;
		if (std::string(argv[i]) == "-OpenCL")
			gpgpuMode = 2;
	}
	if (!gpgpuMode) {
		fprintf(stderr, "Specify the GPGPU backend by \"-<backend>\"\n");
		printf("Available backends: \n");
		printf("\tOpenCL\n");
		#ifdef HAVE_CUDA
		printf("\tCUDA\n");
		#endif
		return 1;
	}

	initGL(argc, argv, (gpgpuMode == 1) ? " (CUDA)" : " (OpenCL)");
	
	#ifdef HAVE_CUDA
	if (gpgpuMode == 1)
		gpgpuCtrl = (gpgpuController*) new cudaController(-1); // -1 == find a device capable of OpenGL interop.
	else
	#endif
		gpgpuCtrl = (gpgpuController*) new openCLController(-1);

	// This texture is filled in GPGPU and used to texture a cube in OpenGL
	glGenTextures(1, &texHandle_gl);
	glEnable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texHandle_gl);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, 512, 512, 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	/*
	   This is how it goes:  1) The OpenGL texture is shared to GPGPU to be used by the GPGPU kernel
							 2) The kernel fills in a buffer based on samples of the texture
							 3) The buffer is written over the texture, and released back to OpenGL
	   (What the kernel actually does is irrelevant, but it essentially incrementally changes the texture)
	*/
	size_t texSize = sizeof(float)*512*512*4;
	try {
		// Allocating the GPGPU resources
		bufHandle_gpgpu = gpgpuCtrl->newBuffer(texSize, NULL); // If not NULL, data is read from the pointer
		texHandle_gpgpu = gpgpuCtrl->makeGLTexture(texHandle_gl);

		// Clearing the texture
		float zero = 0.0f;
		gpgpuCtrl->clearBuffer(bufHandle_gpgpu, &zero);
		gpgpuCtrl->fillTexFromBuffer(texHandle_gpgpu, bufHandle_gpgpu);

		// Kernel sources
		KernelSrc src;
		src.appendFromFile("texfill.gp");
		src.setKeywords(gpgpuCtrl->getKeywords());
		kernelHandle_gpgpu = gpgpuCtrl->newKernel(&src);

		// Binding the buffer and the texture to the kernel
		gpgpuCtrl->setBufferParam(kernelHandle_gpgpu, bufHandle_gpgpu);
		gpgpuCtrl->setTex(kernelHandle_gpgpu, texHandle_gpgpu, "cubetex"); // "Cubetex" == tex name in kernel code

		// Total threads X, block size X; and the same for Y (Y can be omitted for 1D)
		gpgpuCtrl->setKernelExecConf(kernelHandle_gpgpu, 512, 16, 512, 16); 
		gpgpuCtrl->preferSM(kernelHandle_gpgpu); // Prefer as much shared memory as possible (supported by CUDA)

		// Writing out the final sources if you want to check them out
		// (Will be postfixed by ".cu" or ".cl" depending on your backend)
		gpgpuCtrl->writeKernelSrc(kernelHandle_gpgpu, "texfillSource");

		// We can also benchmark the kernel right here and now
		// (Prints statistics to stdout)
		gpgpuCtrl->reportTiming(kernelHandle_gpgpu);

	} catch (std::string e) {
		fprintf(stderr, "Couldn't: \"%s\"\n", e.c_str());
		return 2;
	}

	// Handing control over to GLUT
	glutMainLoop();
}

void gpgpuUpdate() {
	// Executing the kernel
	gpgpuCtrl->executeKernel(kernelHandle_gpgpu);
	
	// Filling in the texture from the buffer data
	gpgpuCtrl->fillTexFromBuffer(texHandle_gpgpu, bufHandle_gpgpu);
}
